"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras.src.datasets.boston_housing import load_data
from keras.src.datasets.california_housing import load_data
